--[[
-- These events can be registered for using the regular CallbackHandler ways.
--
-- "GroupInSpecT_Update", guid, unit, info
-- "GroupInSpecT_Remove, guid
--
-- Where <info> is a table containing some or all of the following:
--   .guid
--   .name
--   .realm
--   .race
--   .race_localized
--   .class
--   .class_localized
--   .class_id
--   .gender -- 2 = male, 3 = female
--   .global_spec_id
--   .spec_index 
--   .spec_name_localized 
--   .spec_description 
--   .spec_icon 
--   .spec_background 
--   .spec_role 
--   .talents = {
--     [<spell_id>] = { 
--       .name
--       .id -- 1 to MAX_NUM_TALENTS
--       .texture
--       .tier
--       .column
--     }
--     ...
--   }
--   .glyphs = {
--     [<spell_id>] = { 
--       .name
--       .glyph_type
--       .id -- 1 to NUM_GLYPH_SLOTS
--       .icon
--       .glyph_id
--     }
--     ...
--   },
--   .lku -- last known unit id
--
-- Functions for external use:
--
--   lib:Rescan ()
--     Force a rescan of all current group members
--
--   lib:QueuedInspections ()
--     Returns an array of GUIDs of outstanding inspects.
--
--   lib:StaleInspections ()
--     Returns an array of GUIDs for which the data has become stale and is
--     awaiting an update (no action required, the refresh happens internally).
--     Due to Blizzard exposing no events on (re/un)talent, there will be
--     frequent marking of inspect data as being stale.
--
--   lib:GetCachedInfo (guid)
--     Returns the cached info for the given GUID, if available, nil otherwise.
--     Information is cached for current group members only.
--
--   lib:GroupUnits ()
--     Returns an array with the set of unit ids for the current group.
--]]

local MAJOR, MINOR = "LibGroupInSpecT-1.0", tonumber (("1.0.0beta2"):match ("(%d+)") or 0)

if not LibStub then error(MAJOR.."requires LibStub") end
local lib, oldversion = LibStub:NewLibrary(MAJOR, MINOR)
if not lib then return end

if lib.state then return end -- Already loaded

LibStub("AceSerializer-3.0"):Embed(lib)
LibStub("AceComm-3.0"):Embed(lib)
local LibCompress = LibStub("LibCompress")
local LibEncode = LibCompress:GetAddonEncodeTable()

local UPDATE_EVENT = "GroupInSpecT_Update"
local REMOVE_EVENT = "GroupInSpecT_Remove"
local COMMS_PREFIX = "LGIST1"

local INSPECT_DELAY = 1.5
local INSPECT_STALE_DELAY = 5
local INSPECT_TIMEOUT = 10 -- If we get no notification within 10s, give up on unit
local me = UnitName("player")


local function debug (...)
--[===[@debug@
  --print (...)
--@end-debug@]===]
end

lib.events = lib.events or LibStub ("CallbackHandler-1.0"):New (lib)

-- Frame for events
local frame = CreateFrame ("Frame", MAJOR .. "_Frame")
lib.frame = frame
frame:Hide()
frame:UnregisterAllEvents ()
frame:RegisterEvent ("PLAYER_LOGIN")
frame:RegisterEvent ("PLAYER_LOGOUT")
frame:SetScript ("OnEvent", function (this, event, ...)
  return lib[event] and lib[event] (lib, ...)
end)

-- Hide our run-state in an easy-to-dump object
lib.state = {
  mainq = {}, staleq = {}, -- inspect queues
  t = 0,
  last_inspect = 0,
  current_guid = nil,
  throttle = 0,
  tt = 0,
  debounce_send_update = 0,
}
lib.cache = {}

-- Get local handles on the key API functions
local CanInspect                      = _G.CanInspect
local ClearInspectPlayer              = _G.ClearInspectPlayer
local GetGlyphSocketInfo              = _G.GetGlyphSocketInfo
local GetNumSubgroupMembers           = _G.GetNumSubgroupMembers
local GetNumSpecializationsForClassID = _G.GetNumSpecializationsForClassID
local GetPlayerInfoByGUID             = _G.GetPlayerInfoByGUID
local GetInspectSpecialization        = _G.GetInspectSpecialization 
local GetSpecialization               = _G.GetSpecialization
local GetSpecializationInfo           = _G.GetSpecializationInfo 
local GetSpecializationInfoForClassID = _G.GetSpecializationInfoForClassID
local GetSpecializationRoleByID       = _G.GetSpecializationRoleByID
local GetSpellInfo                    = _G.GetSpellInfo
local GetTalentInfo                   = _G.GetTalentInfo
local IsInRaid                        = _G.IsInRaid
local NotifyInspect                   = _G.NotifyInspect
local GetNumClasses                   = _G.GetNumClasses
local UnitClass                       = _G.UnitClass
local UnitExists                      = _G.UnitExists
local UnitGUID                        = _G.UnitGUID
local UnitInParty                     = _G.UnitInParty
local UnitInRaid                      = _G.UnitInRaid
local UnitIsConnected                 = _G.UnitIsConnected
local UnitIsPlayer                    = _G.UnitIsPlayer
local UnitIsUnit                      = _G.UnitIsUnit
local UnitName                        = _G.UnitName
  
-- Inspects only work after entering the world, so track that
lib.state.logged_in = IsLoggedIn ()
function lib:PLAYER_LOGIN ()
  lib.state.logged_in = true
  lib:CacheGameData()
  local guid = UnitGUID ("player")
  local info = self:BuildInfo ("player")
  self.events:Fire (UPDATE_EVENT, guid, "player", info)
  frame:RegisterEvent ("INSPECT_READY")
  frame:RegisterEvent ("GROUP_ROSTER_UPDATE")
  frame:RegisterEvent ("UNIT_LEVEL")
  frame:RegisterEvent ("PLAYER_TALENT_UPDATE")
  frame:RegisterEvent ("PLAYER_SPECIALIZATION_CHANGED")
  frame:RegisterEvent ("UNIT_SPELLCAST_SUCCEEDED")
  frame:RegisterEvent ("UNIT_NAME_UPDATE")
  frame:RegisterEvent ("GLYPH_ADDED")
  frame:RegisterEvent ("GLYPH_REMOVED")
end

function lib:PLAYER_LOGOUT ()
  lib.state.logged_in = false
end

-- Simple timer
do
  lib.state.t = 0
  frame:SetScript("OnUpdate", function(this, elapsed)
    lib.state.t = lib.state.t + elapsed
    lib.state.tt = lib.state.tt + elapsed
    local timeout = next (lib.state.mainq) and INSPECT_DELAY or INSPECT_STALE_DELAY
    if lib.state.t > timeout then
      lib:ProcessQueues ()
      lib.state.t = 0
    end
    -- Unthrottle, essentially allowing 1 msg every 3 seconds, but with substantial burst capacity
    if lib.state.tt > 3 and lib.state.throttle > 0 then
      lib.state.throttle = lib.state.throttle - 1
      lib.state.tt = 0
    end
    if lib.state.debounce_send_update > 0 then
      local debounce = lib.state.debounce_send_update - elapsed
      lib.state.debounce_send_update = debounce
      if debounce <= 0 then lib:SendLatestSpecData () end
    end
  end)
end

-- Internal library functions
-- Dridzt: this is in essence static data, so cache them once on login, then do fast local lookups to populate the info
local global_spec_id_to_any = {} -- .spec_index,.spec_name_localized,.spec_description,.spec_icon,.spec_background,.spec_role
local talent_idx_to_spell_id = {}
-- Dridzt: I'd love another way but none of the GetTalent* functions return spellID, GetTalentLink() and parsing the link gives talentID that's not related to spellID as well
-- A quick tooltip scan is cheap though so elegance aside this is a good workaround considering this only runs once
local scanTip = CreateFrame("GameTooltip","LibGroupInSpecTScanTip",nil,"GameTooltipTemplate")
scanTip:SetOwner(UIParent, "ANCHOR_NONE")
function lib:CacheGameData()
  for classID = 1, GetNumClasses() do
    for specIndex = 1, GetNumSpecializationsForClassID(classID) do
      local global_spec_id, name, description, icon, background = GetSpecializationInfoForClassID(classID, specIndex)
      global_spec_id_to_any[global_spec_id] = {}
      global_spec_id_to_any[global_spec_id].spec_index = specIndex
      global_spec_id_to_any[global_spec_id].spec_name_localized = name
      global_spec_id_to_any[global_spec_id].spec_description = description
      global_spec_id_to_any[global_spec_id].spec_icon = icon
      global_spec_id_to_any[global_spec_id].spec_background = background
      global_spec_id_to_any[global_spec_id].spec_role = GetSpecializationRoleByID(global_spec_id) -- TANK, HEALER, DAMAGER roleTokens _G.TANK etc for the localized strings
    end
    talent_idx_to_spell_id[classID] = {}
    for talentIndex=1, MAX_NUM_TALENTS do
      scanTip:ClearLines()
      scanTip:SetTalent(talentIndex,true,nil,nil,classID)
      local spellName, spellRank, spellID = scanTip:GetSpell()
      talent_idx_to_spell_id[classID][talentIndex] = spellID
    end
  end
end

local unitstrings = {
  raid = { "player" }, -- This seems to be needed under certain circumstances. Odd.
  party = { "player" }, -- Player not part of partyN
  player = { "player" }
}
for i = 1,40 do table.insert (unitstrings.raid, "raid"..i) end
for i = 1,4  do table.insert (unitstrings.party, "party"..i) end

-- Returns an array with the set of unit ids for the current group
function lib:GroupUnits ()
  local units
  if IsInRaid () then
    units = unitstrings.raid
  elseif GetNumSubgroupMembers () > 0 then
    units = unitstrings.party
  else
    units = unitstrings.player
  end
  return units
end


function lib:GuidToUnit (guid)
  local info = self.cache[guid]
  if info and info.lku and UnitGUID (info.lku) == guid then return info.lku end

  for i,unit in ipairs (self:GroupUnits ()) do
    if UnitExists (unit) and UnitGUID (unit) == guid then
      if info then info.lku = unit end
      return unit
    end
  end
end


function lib:Query (unit)
  if not UnitIsPlayer (unit) then return end -- NPC

  if UnitIsUnit (unit, "player") then
    self.events:Fire (UPDATE_EVENT, UnitGUID("player"), "player", self:BuildInfo ("player"))
    return
  end

  local mainq, staleq = self.state.mainq, self.state.staleq

  local guid = UnitGUID (unit)
  if not mainq[guid] then
    mainq[guid] = 1
    staleq[guid] = nil
    self.frame:Show () -- Start timer if not already running
  end
end


function lib:Refresh (unit)
  local guid = UnitGUID (unit)
  if not guid then return end
  debug ("Refreshing "..unit)
  if not self.state.mainq[guid] then
    self.state.staleq[guid] = 1
    self.frame:Show ()
  end
end


function lib:ProcessQueues ()
  if not self.state.logged_in then return end
  if InCombatLockdown () then return end -- Never inspect while in combat

  -- Don't' mess with the UI's inspections
  if InspectFrame and InspectFrame:IsShown () then
    return
  end

  local mainq = self.state.mainq
  local staleq = self.state.staleq

  local q = mainq
  if not next (q) then q = staleq end

  if (self.state.last_inspect + INSPECT_TIMEOUT) < GetTime () then
    -- If there was an inspect going, it's timed out so kick it out of the queue
    local guid = self.state.current_guid
    if guid then
      debug ("Inspect timed out for "..guid)
      if not self:GuidToUnit (guid) then
        debug ("No longer applicable, removing from queue")
        mainq[guid], staleq[guid] = nil, nil
      end
      self.state.current_guid = nil
    end
  end

  if self.state.current_guid then return end -- Still waiting on our inspect data

  for guid,_ in pairs (q) do
    local unit = self:GuidToUnit (guid)
    if not unit or not CanInspect (unit) or not UnitIsConnected (unit) then
      q[guid] = nil
    else
      debug ("Inspecting "..unit..", aka ".. (UnitName(unit) or "nil"))
      self.state.current_guid = guid
      NotifyInspect (unit)
      break
    end
  end

  if not next (self.state.mainq) and not next (self.state.staleq) and self.state.throttle == 0 and self.state.debounce_send_update <= 0 then
    frame:Hide() -- Cancel timer, nothing queued and no unthrottling to be done
  end
end


hooksecurefunc("NotifyInspect", function (...) return lib:NotifyInspect (...) end)
function lib:NotifyInspect(unit)
  self.state.last_inspect = GetTime()
end

function lib:UpdatePlayerInfo (guid, unit, info)
  info.class_localized, info.class, info.race_localized, info.race, info.gender, info.name, info.realm = GetPlayerInfoByGUID (guid)
  if info.realm and info.realm == "" then info.realm = nil end
  local _, _, class_id = UnitClass (unit)
  info.class_id = class_id
  info.lku = unit
end

function lib:BuildInfo (unit)
  local guid = UnitGUID (unit)
  if not guid then return end

  local cache = self.cache
  local info = cache[guid] or {}
  cache[guid] = info
  info.guid = guid

  self:UpdatePlayerInfo (guid, unit, info)

  local is_inspect = not UnitIsUnit("player", unit)
  local spec = GetSpecialization ()
  info.global_spec_id = is_inspect and GetInspectSpecialization (unit) or spec and GetSpecializationInfo (spec)
  if info.global_spec_id and not global_spec_id_to_any[info.global_spec_id] then -- not a valid spec_id
    info.global_spec_id = nil
  end
  if info.global_spec_id then
    info.spec_index = global_spec_id_to_any[info.global_spec_id].spec_index
    info.spec_name_localized = global_spec_id_to_any[info.global_spec_id].spec_name_localized
    info.spec_description = global_spec_id_to_any[info.global_spec_id].spec_description
    info.spec_icon = global_spec_id_to_any[info.global_spec_id].spec_icon
    info.spec_background = global_spec_id_to_any[info.global_spec_id].spec_background
    info.spec_role = global_spec_id_to_any[info.global_spec_id].spec_role
  end

  info.talents = info.talents and wipe (info.talents) or {}
  for idx = 1, MAX_NUM_TALENTS do
    local tname, tex, tier, column, sel, avail = GetTalentInfo (idx, is_inspect, nil, unit, info.class_id)
    local spellID = talent_idx_to_spell_id[info.class_id][idx]
    if not info.talents[spellID] and sel then
      info.talents[spellID] = {}
      info.talents[spellID].name = tname
      info.talents[spellID].texture = tex
      info.talents[spellID].tier = tier
      info.talents[spellID].column = column
      info.talents[spellID].id = idx
    end
  end
  info.glyphs = info.glyphs and wipe (info.glyphs) or {}
  for idx = 1,NUM_GLYPH_SLOTS do
    local enabled, glyphType, glyphTooltipIndex, glyphSpell, iconFilename, glyphID = GetGlyphSocketInfo (idx, nil, is_inspect, unit)
    if not info.glyphs[glyphSpell] and enabled and glyphID then
      info.glyphs[glyphSpell] = {}
      info.glyphs[glyphSpell].glyph_type = glyphType
      info.glyphs[glyphSpell].id = idx
      info.glyphs[glyphSpell].icon = iconFilename
      info.glyphs[glyphSpell].glyph_id = glyphID
      info.glyphs[glyphSpell].name = glyphSpell and (GetSpellInfo(glyphSpell)) or nil
    end
  end
  return info
end

function lib:INSPECT_READY (guid)
  local unit = self:GuidToUnit (guid)
  local finalize = false
  if unit then
    if guid == self.state.current_guid then
      self.state.current_guid = nil -- Got what we asked for
      finalize = true
      debug("Got inspection data for requested guid "..guid)
    end

    local mainq, staleq = self.state.mainq, self.state.staleq
    mainq[guid], staleq[guid] = nil, nil

    local gspec_id = GetInspectSpecialization (unit)
    if not global_spec_id_to_any[gspec_id] then -- Bah, got garbage, flag as stale and try again
      self.state.staleq[guid] = guid
      return
    end

    self.events:Fire (UPDATE_EVENT, guid, unit, self:BuildInfo (unit))
  end
  if finalize then
    ClearInspectPlayer ()
  end
end


-- Group handling parts

function lib:GROUP_ROSTER_UPDATE ()
  local group = self.cache
  local members = {}
  local units = self:GroupUnits ()
  -- Find new members
  for i,unit in ipairs (self:GroupUnits ()) do
    local guid = UnitGUID (unit)
    if guid then
      members[guid] = 1
      if not group[guid] then
        self:Query (unit)
        -- Update with what we have so far (guid, unit, name/class/race?)
        self.events:Fire (UPDATE_EVENT, guid, unit, self:BuildInfo (unit))
      end
    end
  end
  -- Find removed members
  for guid,info in pairs (group) do
    if not members[guid] then
      group[guid] = nil
      self.events:Fire (REMOVE_EVENT, guid, nil)
    end
  end
end

function lib:DoPlayerUpdate ()
  self:Query ("player")
  self.state.debounce_send_update = 2.5 -- Hold off 2.5sec before sending update
  frame:Show ()
end

function lib:SendLatestSpecData () 
  local guid = UnitGUID ("player")
  local info = self.cache[guid]
  if not info then return end
  
  local serializedinfo = lib:Serialize(info)
  local compressedinfo = LibCompress:CompressHuffman(serializedinfo)
  local encodedinfo = LibEncode:Encode(compressedinfo)
  
  if UnitInBattleground ("player") then
    debug("Sending LGIST update to BATTLEGROUND")
    lib:SendCommMessage(COMMS_PREFIX, encodedinfo, "BATTLEGROUND")
  elseif IsInRaid () and GetNumGroupMembers () > 0 then
    debug("Sending LGIST update to RAID")
    lib:SendCommMessage(COMMS_PREFIX, encodedinfo, "RAID")
  end

end

function lib:OnCommReceived (prefix, datastr, scope, sender)
  if prefix ~= COMMS_PREFIX then return end
  debug ("Incoming LGIST update from "..(scope or "nil").."/"..(sender or "nil")..": "..(datastr or "nil"))
  if sender == me then return end

  self.state.throttle = self.state.throttle + 1
  frame:Show () -- Ensure we're unthrottling
  if self.state.throttle > 40 then return end -- If we ever hit this, someone's being "funny"
  
  local decodedinfo = LibEncode:Decode(datastr)
  local decompressedinfo, decomp_error = LibCompress:Decompress(decodedinfo)
  if not decompressedinfo then
    return debug ("decompression", decomp_error)
  end
  local success, receivedinfo = lib:Deserialize(decompressedinfo)
  if not success then
    return debug ("deserialization", receivedinfo)
  end
  
  local guid = receivedinfo.guid
  local info = guid and self.cache[guid]
  if not info then return end
  
  local unit = self:GuidToUnit (guid)
  if UnitIsUnit( unit, "player" ) then return end -- we're already up-to-date
  
  if unit and guid then
    self.events:Fire (UPDATE_EVENT, guid, unit, receivedinfo)
  end
end
lib:RegisterComm(COMMS_PREFIX)

function lib:UNIT_LEVEL (unit)
  if UnitInRaid (unit) or UnitInParty (unit) then
    self:Refresh (unit)
  end
  if UnitIsUnit (unit, "player") then
    self:DoPlayerUpdate ()
  end
end

function lib:PLAYER_TALENT_UPDATE ()
  self:DoPlayerUpdate ()
end

function lib:PLAYER_SPECIALIZATION_CHANGED (unit)
--  This event seems to fire a lot, and for no particular reason *sigh*
--  if UnitInRaid (unit) or UnitInParty (unit) then
--    self:Refresh (unit)
--  end
  if unit and UnitIsUnit (unit, "player") then
    self:DoPlayerUpdate ()
  end
end

function lib:GLYPH_ADDED ()
  self:DoPlayerUpdate ()
end

function lib:GLYPH_REMOVED ()
  self:DoPlayerUpdate ()
end

function lib:UNIT_NAME_UPDATE (unit)
  local group = self.cache
  local guid = UnitGUID (unit)
  local info = guid and group[guid]
  if info then
    self:UpdatePlayerInfo (guid, unit, info)
    if info.name ~= UNKNOWN then
      self.events:Fire (UPDATE_EVENT, guid, unit, info)
    end
  end
end

local dual_spec_spells = {}
for i, spellid in ipairs (TALENT_ACTIVATION_SPELLS) do dual_spec_spells[spellid] = true end
function lib:UNIT_SPELLCAST_SUCCEEDED (unit, spellname, rank, lineid, spellid)
  if dual_spec_spells[spellid] then
    self:Query (unit) -- Definitely changed, so high prio refresh
  end
end


-- External library functions

function lib:QueuedInspections ()
  local q = {}
  for guid,_ in pairs (self.state.mainq) do
    table.insert (q, guid)
  end
  return q
end

function lib:StaleInspections ()
  local q = {}
  for guid,_ in pairs (self.state.staleq) do
    table.insert (q, guid)
  end
  return q
end

function lib:GetCachedInfo (guid)
  local group = self.cache
  return guid and group[guid]
end

function lib:Rescan ()
  local q = self.state.mainq
  for i,unit in ipairs (self:GroupUnits ()) do
    if UnitExists (unit) then
      local guid = UnitGUID (unit)
      if guid then q[guid] = 1 end
    end
  end
  self.frame:Show () -- Start timer if not already running

  -- Evict any stale entries
  self:GROUP_ROSTER_UPDATE ()
end


--[[
Notes:
MAX_NUM_TALENT_TIERS = 6;
NUM_TALENT_COLUMNS = 3;
MAX_TALENT_GROUPS = 2;
MAX_TALENT_TABS = 4;
MAX_NUM_TALENTS = 18;
local ClassID = 1 to GetNumClasses()
local class, classTag, ClassID = GetClassInfo(ClassID)
local global_spec_id, name, description, icon, background = GetSpecializationInfoByID(global_spec_id)
local specIndex = 1 to GetNumSpecializationsForClassID(class_id)
local global_spec_id, name, description, icon, background = GetSpecializationInfoForClassID(class_id, specIndex)
local roleToken = GetSpecializationRoleByID(global_spec_id)
local numTalents = GetNumTalents()
local name, texture, tier, column, selected, available = GetTalentInfo(1,true,nil,nil,11) -- talentNum, isInspect, talentGroup, unit, ClassID
local link = GetTalentLink(1,true,11); -- self:GetID(), InspectTalentFrame.InspectTalents.inspect, classID
GameTooltip:SetTalent(1,true,nil,nil,11); -- self:GetID(),true, self.talentGroup, INSPECTED_UNIT, classID
local spellName, spellRank, spellID = GameTooltip:GetSpell()
local enabled, glyphType, glyphTooltipIndex, glyphSpell, iconFilename, glyphID = GetGlyphSocketInfo(id, talentGroup, true, INSPECTED_UNIT); -- id 1 to NUM_GLYPH_SLOTS, nil, true
-- enabled, GLYPH_TYPE_MINOR|GLYPH_TYPE_MAJOR, 1-6, spellID, iconTexture, glyphID
]]